@@ -2,7 +2,7 @@
  *
  * rewriteManip.c
  *
- * Portions Copyright (c) 1996-2016, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2018, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *
@@ -1000,7 +1000,7 @@ AddQual(Query *parsetree, Node *qual)
 		else
 			ereport(ERROR,
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-			  errmsg("conditional utility statements are not implemented")));
+					 errmsg("conditional utility statements are not implemented")));
 	}
 
 	if (parsetree->setOperations != NULL)
@@ -1143,7 +1143,7 @@ replace_rte_variables_mutator(Node *node,
 			/* Found a matching variable, make the substitution */
 			Node	   *newnode;
 
-			newnode = (*context->callback) (var, context);
+			newnode = context->callback(var, context);
 			/* Detect if we are adding a sublink to query */
 			if (!context->inserted_sublink)
 				context->inserted_sublink = checkExprHasSubLink(newnode);
@@ -1166,7 +1166,7 @@ replace_rte_variables_mutator(Node *node,
 			 */
 			ereport(ERROR,
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-				   errmsg("WHERE CURRENT OF on a view is not implemented")));
+					 errmsg("WHERE CURRENT OF on a view is not implemented")));
 		}
 		/* otherwise fall through to copy the expr normally */
 	}
@@ -1203,14 +1203,14 @@ replace_rte_variables_mutator(Node *node,
  * appear in the expression.
  *
  * If the expression tree contains a whole-row Var for the target RTE,
- * the Var is not changed but *found_whole_row is returned as TRUE.
- * For most callers this is an error condition, but we leave it to the caller
- * to report the error so that useful context can be provided.  (In some
- * usages it would be appropriate to modify the Var's vartype and insert a
- * ConvertRowtypeExpr node to map back to the original vartype.  We might
- * someday extend this function's API to support that.  For now, the only
- * concession to that future need is that this function is a tree mutator
- * not just a walker.)
+ * *found_whole_row is set to true.  In addition, if to_rowtype is
+ * not InvalidOid, we replace the Var with a Var of that vartype, inserting
+ * a ConvertRowtypeExpr to map back to the rowtype expected by the expression.
+ * (Therefore, to_rowtype had better be a child rowtype of the rowtype of the
+ * RTE we're changing references to.)  Callers that don't provide to_rowtype
+ * should report an error if *found_row_type is true; we don't do that here
+ * because we don't know exactly what wording for the error message would
+ * be most appropriate.  The caller will be aware of the context.
  *
  * This could be built using replace_rte_variables and a callback function,
  * but since we don't ever need to insert sublinks, replace_rte_variables is
@@ -1223,6 +1223,7 @@ typedef struct
 	int			sublevels_up;	/* (current) nesting depth */
 	const AttrNumber *attno_map;	/* map array for user attnos */
 	int			map_length;		/* number of entries in attno_map[] */
+	Oid			to_rowtype;		/* change whole-row Vars to this type */
 	bool	   *found_whole_row;	/* output flag */
 } map_variable_attnos_context;
 
@@ -1243,7 +1244,8 @@ map_variable_attnos_mutator(Node *node,
 			Var		   *newvar = (Var *) palloc(sizeof(Var));
 			int			attno = var->varattno;
 
-			*newvar = *var;
+			*newvar = *var;		/* initially copy all fields of the Var */
+
 			if (attno > 0)
 			{
 				/* user-defined column, replace attno */
@@ -1257,11 +1259,77 @@ map_variable_attnos_mutator(Node *node,
 			{
 				/* whole-row variable, warn caller */
 				*(context->found_whole_row) = true;
+
+				/* If the caller expects us to convert the Var, do so. */
+				if (OidIsValid(context->to_rowtype) &&
+					context->to_rowtype != var->vartype)
+				{
+					ConvertRowtypeExpr *r;
+
+					/* This certainly won't work for a RECORD variable. */
+					Assert(var->vartype != RECORDOID);
+
+					/* Var itself is changed to the requested type. */
+					newvar->vartype = context->to_rowtype;
+
+					/*
+					 * Add a conversion node on top to convert back to the
+					 * original type expected by the expression.
+					 */
+					r = makeNode(ConvertRowtypeExpr);
+					r->arg = (Expr *) newvar;
+					r->resulttype = var->vartype;
+					r->convertformat = COERCE_IMPLICIT_CAST;
+					r->location = -1;
+
+					return (Node *) r;
+				}
 			}
 			return (Node *) newvar;
 		}
 		/* otherwise fall through to copy the var normally */
 	}
+	else if (IsA(node, ConvertRowtypeExpr))
+	{
+		ConvertRowtypeExpr *r = (ConvertRowtypeExpr *) node;
+		Var		   *var = (Var *) r->arg;
+
+		/*
+		 * If this is coercing a whole-row Var that we need to convert, then
+		 * just convert the Var without adding an extra ConvertRowtypeExpr.
+		 * Effectively we're simplifying var::parenttype::grandparenttype into
+		 * just var::grandparenttype.  This avoids building stacks of CREs if
+		 * this function is applied repeatedly.
+		 */
+		if (IsA(var, Var) &&
+			var->varno == context->target_varno &&
+			var->varlevelsup == context->sublevels_up &&
+			var->varattno == 0 &&
+			OidIsValid(context->to_rowtype) &&
+			context->to_rowtype != var->vartype)
+		{
+			ConvertRowtypeExpr *newnode;
+			Var		   *newvar = (Var *) palloc(sizeof(Var));
+
+			/* whole-row variable, warn caller */
+			*(context->found_whole_row) = true;
+
+			*newvar = *var;		/* initially copy all fields of the Var */
+
+			/* This certainly won't work for a RECORD variable. */
+			Assert(var->vartype != RECORDOID);
+
+			/* Var itself is changed to the requested type. */
+			newvar->vartype = context->to_rowtype;
+
+			newnode = (ConvertRowtypeExpr *) palloc(sizeof(ConvertRowtypeExpr));
+			*newnode = *r;		/* initially copy all fields of the CRE */
+			newnode->arg = (Expr *) newvar;
+
+			return (Node *) newnode;
+		}
+		/* otherwise fall through to process the expression normally */
+	}
 	else if (IsA(node, Query))
 	{
 		/* Recurse into RTE subquery or not-yet-planned sublink subquery */
@@ -1283,7 +1351,7 @@ Node *
 map_variable_attnos(Node *node,
 					int target_varno, int sublevels_up,
 					const AttrNumber *attno_map, int map_length,
-					bool *found_whole_row)
+					Oid to_rowtype, bool *found_whole_row)
 {
 	map_variable_attnos_context context;
 
@@ -1291,6 +1359,7 @@ map_variable_attnos(Node *node,
 	context.sublevels_up = sublevels_up;
 	context.attno_map = attno_map;
 	context.map_length = map_length;
+	context.to_rowtype = to_rowtype;
 	context.found_whole_row = found_whole_row;
 
 	*found_whole_row = false;
@@ -1397,12 +1466,12 @@ ReplaceVarsFromTargetList_callback(Var *var,
 				 */
 				return coerce_to_domain((Node *) makeNullConst(var->vartype,
 															   var->vartypmod,
-															 var->varcollid),
+															   var->varcollid),
 										InvalidOid, -1,
 										var->vartype,
+										COERCION_IMPLICIT,
 										COERCE_IMPLICIT_CAST,
 										-1,
-										false,
 										false);
 		}
 		elog(ERROR, "could not find replacement targetlist entry for attno %d",
@@ -1412,11 +1481,11 @@ ReplaceVarsFromTargetList_callback(Var *var,
 	else
 	{
 		/* Make a copy of the tlist item to return */
-		Node	   *newnode = copyObject(tle->expr);
+		Expr	   *newnode = copyObject(tle->expr);
 
 		/* Must adjust varlevelsup if tlist item is from higher query */
 		if (var->varlevelsup > 0)
-			IncrementVarSublevelsUp(newnode, var->varlevelsup, 0);
+			IncrementVarSublevelsUp((Node *) newnode, var->varlevelsup, 0);
 
 		/*
 		 * Check to see if the tlist item contains a PARAM_MULTIEXPR Param,
@@ -1428,12 +1497,12 @@ ReplaceVarsFromTargetList_callback(Var *var,
 		 * create semantic oddities that users of rules would probably prefer
 		 * not to cope with.  So treat it as an unimplemented feature.
 		 */
-		if (contains_multiexpr_param(newnode, NULL))
+		if (contains_multiexpr_param((Node *) newnode, NULL))
 			ereport(ERROR,
 					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
 					 errmsg("NEW variables in ON UPDATE rules cannot reference columns that are part of a multiple assignment in the subject UPDATE command")));
 
-		return newnode;
+		return (Node *) newnode;
 	}
 }
 
